##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from zope.authentication.interfaces import IAuthentication
from zope.authentication.interfaces import PrincipalLookupError
from zope.interface import implementer
from zope.security.interfaces import IPrincipal


@implementer(IAuthentication)
class DummyPrincipalRegistry:
    """Dummy principal registry that only implements getPrincipal
    and definePrincipal method that are needed for securitypolicy tests."""

    def __init__(self):
        self._principals = {}

    def getPrincipal(self, id):
        if id not in self._principals:
            raise PrincipalLookupError(id)
        return self._principals[id]

    def definePrincipal(self, id, title='', description=''):
        p = DummyPrincipal(id, title, description)
        self._principals[id] = p
        return p


principalRegistry = DummyPrincipalRegistry()


@implementer(IPrincipal)
class DummyPrincipal:
    """Very simple principal implementation"""

    def __init__(self, id, title='', description=''):
        self.id = id
        self.title = title
        self.description = description



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Security Settings Tests
"""
import pickle
import unittest

from zope.securitypolicy.interfaces import Allow


class Test(unittest.TestCase):

    def testPickleUnpickle(self):
        for prot in range(0, pickle.HIGHEST_PROTOCOL):
            s = pickle.dumps(Allow, prot)
            newAllow = pickle.loads(s)

            self.assertIs(newAllow, Allow)

    def testDescription(self):
        self.assertEqual("Explicit allow setting for permissions",
                         Allow.getDescription())

    def testName(self):
        self.assertEqual("Allow",
                         Allow.getName())

    def testStr(self):
        self.assertEqual("PermissionSetting: Allow",
                         str(Allow))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests the zope policy.
"""

import unittest
from doctest import DocFileSuite

from zope.annotation.attribute import AttributeAnnotations
from zope.annotation.interfaces import IAnnotatable
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.component import provideAdapter
from zope.component.testing import setUp as componentSetUp
from zope.component.testing import tearDown as componentTearDown
from zope.security.management import endInteraction
from zope.testing.cleanup import CleanUp

from zope import interface
from zope.securitypolicy import zopepolicy
from zope.securitypolicy.grantinfo import AnnotationGrantInfo
from zope.securitypolicy.interfaces import IGrantInfo
from zope.securitypolicy.interfaces import IPrincipalPermissionManager
from zope.securitypolicy.interfaces import IPrincipalRoleManager
from zope.securitypolicy.interfaces import IRolePermissionManager
from zope.securitypolicy.principalpermission import \
    AnnotationPrincipalPermissionManager
from zope.securitypolicy.principalrole import AnnotationPrincipalRoleManager
from zope.securitypolicy.rolepermission import AnnotationRolePermissionManager


class TestZCML(CleanUp, unittest.TestCase):

    def testMetaZCML(self):
        import zope.configuration

        import zope.securitypolicy
        zope.configuration.xmlconfig.file("meta.zcml", zope.securitypolicy)

    def testConfigureZCML(self):
        import zope.configuration

        import zope.securitypolicy
        zope.configuration.xmlconfig.file(
            "configure.zcml", zope.securitypolicy)

    def testSecuritypolicyZCML(self):
        import zope.annotation
        import zope.configuration

        import zope.securitypolicy
        zope.configuration.xmlconfig.file(
            "configure.zcml", zope.annotation)

        zope.configuration.xmlconfig.file(
            "securitypolicy.zcml", zope.securitypolicy)

        settings = zopepolicy.settingsForObject(self)
        self.assertEqual(
            settings[0],
            ('(no name)', {})
        )
        self.assertEqual(
            settings[1][0],
            'global settings'
        )

        self.assertIn(
            'principalPermissions',
            settings[1][1]
        )
        self.assertIn(
            'rolePermissions',
            settings[1][1]
        )
        self.assertIn(
            'principalRoles',
            settings[1][1]
        )

        # Making us annotatable changes our data; we don't have anything
        # but we do have managers
        interface.alsoProvides(self, IAttributeAnnotatable)
        settings = zopepolicy.settingsForObject(self)
        self.assertEqual(
            settings[0], ('(no name)', {
                'principalPermissions': [],
                'principalRoles': [],
                'rolePermissions': []}))


class TestZopePolicy(CleanUp, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.policy = zopepolicy.ZopeSecurityPolicy()

    def test_checkPermission_system_user(self):
        from zope.security.management import system_user

        class Participation:
            principal = system_user
            interaction = None

        self.policy.add(Participation())

        self.assertTrue(self.policy.checkPermission('perm', self))

    def test_checkPermission_multiple_participations_for_same_id(self):

        class Principal:
            id = 'principal'

        class Participation:
            principal = Principal()
            interaction = None

        self.policy.add(Participation())
        self.policy.add(Participation())

        invoked_counter = []

        def cached_decision(self, *args):
            invoked_counter.append(args)
            return True

        self.policy.cached_decision = cached_decision
        self.assertTrue(self.policy.checkPermission('perm', self))
        self.assertEqual(1, len(invoked_counter))

    def test__findGroupsFor_seen(self):
        group_id = 'group'

        class Principal:
            groups = (group_id,)

        seen = {group_id}

        # Does nothing because we've already been seen
        self.assertEqual(
            self.policy._findGroupsFor(Principal(), None, seen),
            ()
        )

    def test__findGroupsFor_LookupError(self):
        # lookup errors are ignored
        from zope.authentication.interfaces import PrincipalLookupError

        class Principal:
            groups = ('group',)

        def getPrincipal(gid):
            raise PrincipalLookupError(gid)

        self.assertEqual(
            self.policy._findGroupsFor(Principal(), getPrincipal, []),
            ()
        )


def setUp(test):
    componentSetUp()
    endInteraction()
    provideAdapter(AttributeAnnotations)
    provideAdapter(AnnotationPrincipalPermissionManager, (IAnnotatable,),
                   IPrincipalPermissionManager)
    provideAdapter(AnnotationPrincipalRoleManager, (IAnnotatable,),
                   IPrincipalRoleManager)
    provideAdapter(AnnotationRolePermissionManager, (IAnnotatable,),
                   IRolePermissionManager)
    provideAdapter(AnnotationGrantInfo, (IAnnotatable,), IGrantInfo)


def test_suite():
    return unittest.TestSuite((
        DocFileSuite('zopepolicy.txt',
                     package='zope.securitypolicy',
                     setUp=setUp, tearDown=componentTearDown),
        unittest.defaultTestLoader.loadTestsFromName(__name__)
    ))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test handler for Annotation Role Permission Manager.
"""
import unittest

from zope.annotation.attribute import AttributeAnnotations
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.component.testing import PlacelessSetup
from zope.interface import implementer
from zope.security.interfaces import IPermission
from zope.security.permission import Permission

from zope.securitypolicy.interfaces import Allow
from zope.securitypolicy.interfaces import Deny
from zope.securitypolicy.interfaces import IRole
from zope.securitypolicy.interfaces import Unset
from zope.securitypolicy.role import Role
from zope.securitypolicy.rolepermission import AnnotationRolePermissionManager


@implementer(IAttributeAnnotatable)
class Manageable:
    pass


class Test(PlacelessSetup, unittest.TestCase):

    def setUp(self):
        PlacelessSetup.setUp(self)
        provideAdapter(AttributeAnnotations)

        read = Permission('read', 'Read Something')
        provideUtility(read, IPermission, read.id)
        self.read = read.id

        write = Permission('write', 'Write Something')
        provideUtility(write, IPermission, write.id)
        self.write = write.id

        peon = Role('peon', 'Poor Slob')
        provideUtility(peon, IRole, peon.id)
        self.peon = peon.id

        manager = Role('manager', 'Supreme Being')
        provideUtility(manager, IRole, manager.id)
        self.manager = manager.id

    def testNormal(self):
        obj = Manageable()
        mgr = AnnotationRolePermissionManager(obj)
        mgr.grantPermissionToRole(self.read, self.manager)
        mgr.grantPermissionToRole(self.write, self.manager)
        mgr.grantPermissionToRole(self.write, self.manager)

        mgr.grantPermissionToRole(self.read, self.peon)

        l_ = list(mgr.getPermissionsForRole(self.manager))
        self.assertIn((self.read, Allow), l_)
        self.assertIn((self.write, Allow), l_)

        l_ = list(mgr.getPermissionsForRole(self.peon))
        self.assertEqual([(self.read, Allow)], l_)

        l_ = list(mgr.getRolesForPermission(self.read))
        self.assertIn((self.manager, Allow), l_)
        self.assertIn((self.peon, Allow), l_)

        l_ = list(mgr.getRolesForPermission(self.write))
        self.assertEqual(l_, [(self.manager, Allow)])

        mgr.denyPermissionToRole(self.read, self.peon)
        l_ = list(mgr.getPermissionsForRole(self.peon))
        self.assertEqual(l_, [(self.read, Deny)])

        mgr.unsetPermissionFromRole(self.read, self.peon)

        l_ = list(mgr.getRolesForPermission(self.read))
        self.assertEqual(l_, [(self.manager, Allow)])

        self.assertEqual(mgr.getSetting(self.read, self.peon), Unset)
        self.assertEqual(mgr.getSetting(self.read, self.peon, 1), 1)



##############################################################################
#
# Copyright (c) 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from zope.annotation.interfaces import IAnnotations

from zope import interface
from zope.securitypolicy import grantinfo


# pylint:disable=protected-access

class Manager:

    def __init__(self):
        self._bycol = {}
        self._byrow = {}


@interface.implementer(IAnnotations)
class Annotated(dict):
    "Annotations"


class TestAnnotationGrantInfo(unittest.TestCase):

    def _makeContext(self):
        # Return a populated context with all three role managers
        prinper = Manager()
        prinrole = Manager()
        roleperm = Manager()

        context = Annotated()
        context[grantinfo.prinperkey] = prinper
        context[grantinfo.prinrolekey] = prinrole
        context[grantinfo.rolepermkey] = roleperm

        return context, prinper, prinrole, roleperm

    def _makeOne(self, context):
        return grantinfo.AnnotationGrantInfo(context)

    def test_no_annotations(self):
        # If we provide a context that has no IAnnotations,
        # we get defaults

        info = self._makeOne(None)

        self.assertFalse(info)
        self.assertEqual(grantinfo.Unset,
                         info.principalPermissionGrant(None, None))
        self.assertEqual([],
                         info.getRolesForPermission(None))
        self.assertEqual([],
                         info.getRolesForPrincipal(None))

    def test_manager_attributes(self):
        context, prinper, prinrole, roleperm = self._makeContext()
        info = self._makeOne(context)

        self.assertIs(info.prinper, prinper._bycol)
        self.assertIs(info.prinrole, prinrole._bycol)
        self.assertIs(info.permrole, roleperm._byrow)

    def test_principal_permission_grant(self):
        context, prinper, _, _ = self._makeContext()
        grants = prinper._bycol["principal"] = {}
        grant = grants["permission"] = object()

        info = self._makeOne(context)

        self.assertIs(grant,
                      info.principalPermissionGrant("principal", "permission"))

        self.assertIs(
            grantinfo.Unset,
            info.principalPermissionGrant(
                "principal",
                "other permission"))

        self.assertIs(
            grantinfo.Unset,
            info.principalPermissionGrant(
                "other principal",
                "other permission"))

    def test_roles_for_permission(self):
        context, _, _, roleperm = self._makeContext()
        grants = roleperm._byrow["permission"] = {}
        grants['key'] = 'value'

        info = self._makeOne(context)

        self.assertEqual([('key', 'value')],
                         info.getRolesForPermission('permission'))
        self.assertEqual([],
                         info.getRolesForPermission("other permission"))

    def test_roles_for_principal(self):
        context, _, prinper, _ = self._makeContext()
        grants = prinper._bycol["principal"] = {}
        grants['key'] = 'value'

        info = self._makeOne(context)

        self.assertEqual([('key', 'value')],
                         info.getRolesForPrincipal('principal'))
        self.assertEqual([],
                         info.getRolesForPrincipal("other principal"))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
#############################################################################
"""Test SecurityMap implementations
"""
import unittest

from zope.security.management import endInteraction
from zope.security.management import getInteraction
from zope.security.management import newInteraction
from zope.security.management import setSecurityPolicy

from zope.securitypolicy.securitymap import PersistentSecurityMap
from zope.securitypolicy.securitymap import SecurityMap


class InteractionStub:
    invalidated = 0

    def invalidate_cache(self):
        self.invalidated += 1


class TestSecurityMap(unittest.TestCase):

    def setUp(self):
        self.oldpolicy = setSecurityPolicy(InteractionStub)
        newInteraction()

    def tearDown(self):
        endInteraction()
        setSecurityPolicy(self.oldpolicy)

    def _getSecurityMap(self):
        return SecurityMap()

    def test_addCell(self):
        map = self._getSecurityMap()
        self.assertEqual(getInteraction().invalidated, 0)
        map.addCell(0, 0, 'aa')
        self.assertEqual(getInteraction().invalidated, 1)
        self.assertEqual(map._byrow[0][0], 'aa')
        self.assertEqual(map._bycol[0][0], 'aa')

        map.addCell(1, 0, 'ba')
        self.assertEqual(getInteraction().invalidated, 2)
        self.assertEqual(map._byrow[1][0], 'ba')
        self.assertEqual(map._bycol[0][1], 'ba')

        map.addCell(5, 3, 'fd')
        self.assertEqual(getInteraction().invalidated, 3)
        self.assertEqual(map._byrow[5][3], 'fd')
        self.assertEqual(map._bycol[3][5], 'fd')

    def test_addCell_no_invalidation(self):

        class NoInvalidation:
            attrs = ()

            def __getattr__(self, name):
                self.attrs += (name,)
                return object.__getattr__(self, name)

        setSecurityPolicy(NoInvalidation)
        endInteraction()
        newInteraction()

        map = self._getSecurityMap()
        map.addCell(0, 0, 'aa')
        self.assertIn('invalidate_cache', getInteraction().attrs)

    def test_addCell_noninteger(self):
        map = self._getSecurityMap()
        map.addCell(0.3, 0.4, 'entry')
        self.assertEqual(map._byrow[0.3][0.4], 'entry')
        self.assertEqual(map._bycol[0.4][0.3], 'entry')

        marker = object()
        map.addCell('a', 'b', marker)
        self.assertEqual(map._byrow['a']['b'], marker)
        self.assertEqual(map._bycol['b']['a'], marker)

    def test_delCell(self):
        map = self._getSecurityMap()
        self.assertEqual(getInteraction().invalidated, 0)
        map._byrow[0] = {}
        map._bycol[1] = {}
        map._byrow[0][1] = 'aa'
        map._bycol[1][0] = 'aa'
        map.delCell(0, 1)
        self.assertEqual(getInteraction().invalidated, 1)
        self.assertEqual(map._byrow.get(0), None)
        self.assertEqual(map._bycol.get(1), None)

    def test_queryCell(self):
        map = self._getSecurityMap()
        map._byrow[0] = {}
        map._bycol[1] = {}
        map._byrow[0][1] = 'aa'
        map._bycol[1][0] = 'aa'

        marker = object()
        self.assertEqual(map.queryCell(0, 1), 'aa')
        self.assertEqual(map.queryCell(1, 0), None)
        self.assertEqual(map.queryCell(1, 0, marker), marker)

    def test_getCell(self):
        map = self._getSecurityMap()
        map._byrow[0] = {}
        map._bycol[1] = {}
        map._byrow[0][1] = 'aa'
        map._bycol[1][0] = 'aa'

        self.assertEqual(map.getCell(0, 1), 'aa')
        self.assertRaises(KeyError, map.getCell, 1, 0)

    def test_getRow(self):
        map = self._getSecurityMap()
        map._byrow[0] = {}
        map._byrow[0][1] = 'ab'
        map._byrow[0][2] = 'ac'
        map._byrow[1] = {}
        map._byrow[1][1] = 'bb'
        map._bycol[1] = {}
        map._bycol[1][0] = 'ab'
        map._bycol[1][1] = 'bb'
        map._bycol[2] = {}
        map._bycol[2][0] = 'ac'

        self.assertEqual(map.getRow(0), [(1, 'ab'), (2, 'ac')])
        self.assertEqual(map.getRow(1), [(1, 'bb')])
        self.assertEqual(map.getRow(2), [])

    def test_getCol(self):
        map = self._getSecurityMap()
        map._byrow[0] = {}
        map._byrow[0][1] = 'ab'
        map._byrow[0][2] = 'ac'
        map._byrow[1] = {}
        map._byrow[1][1] = 'bb'
        map._bycol[1] = {}
        map._bycol[1][0] = 'ab'
        map._bycol[1][1] = 'bb'
        map._bycol[2] = {}
        map._bycol[2][0] = 'ac'

        self.assertEqual(map.getCol(1), [(0, 'ab'), (1, 'bb')])
        self.assertEqual(map.getCol(2), [(0, 'ac')])
        self.assertEqual(map.getCol(0), [])

    def test_getAllCells(self):
        map = self._getSecurityMap()
        map._byrow[0] = {}
        map._byrow[0][1] = 'ab'
        map._byrow[0][2] = 'ac'
        map._byrow[1] = {}
        map._byrow[1][1] = 'bb'
        map._bycol[1] = {}
        map._bycol[1][0] = 'ab'
        map._bycol[1][1] = 'bb'
        map._bycol[2] = {}
        map._bycol[2][0] = 'ac'

        self.assertEqual(map.getCol(1), [(0, 'ab'), (1, 'bb')])
        self.assertEqual(map.getCol(2), [(0, 'ac')])
        self.assertEqual(map.getCol(0), [])


class TestPersistentSecurityMap(TestSecurityMap):

    def _getSecurityMap(self):
        return PersistentSecurityMap()


class TestAnnotationSecurityMap(unittest.TestCase):

    def test_changed_sets_map(self):
        from zope.annotation.interfaces import IAnnotations

        from zope.securitypolicy.securitymap import AnnotationSecurityMap

        class Context:
            def __init__(self):
                self.annotations = {}

            def __conform__(self, iface):
                if iface is IAnnotations:
                    return self.annotations

        class ASM(AnnotationSecurityMap):
            # 'key' is expected to be defined by subclasses
            key = 'key'

        context = Context()
        sec_map = ASM(context)
        # No key added yet.
        self.assertEqual(context.annotations, {})
        self.assertIsNone(sec_map.map)

        # Adding a cell sets the map
        sec_map.addCell('row', 'col', 'val')

        self.assertEqual(len(context.annotations), 1)
        self.assertIn(ASM.key, context.annotations)
        psm = context.annotations[ASM.key]
        self.assertIs(psm, sec_map.map)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Define Zope's default security policy
"""

import zope.interface
from zope.authentication.interfaces import IAuthentication
from zope.authentication.interfaces import PrincipalLookupError
from zope.component import getUtility
from zope.security.checker import CheckerPublic
from zope.security.interfaces import ISecurityPolicy
from zope.security.management import system_user
from zope.security.proxy import removeSecurityProxy
from zope.security.simplepolicies import ParanoidSecurityPolicy

from zope.securitypolicy.interfaces import Allow
from zope.securitypolicy.interfaces import Deny
from zope.securitypolicy.interfaces import IPrincipalPermissionMap
from zope.securitypolicy.interfaces import IPrincipalRoleMap
from zope.securitypolicy.interfaces import IRolePermissionMap
from zope.securitypolicy.interfaces import Unset
from zope.securitypolicy.principalpermission import principalPermissionManager
from zope.securitypolicy.principalrole import principalRoleManager
from zope.securitypolicy.rolepermission import rolePermissionManager


globalPrincipalPermissionSetting = principalPermissionManager.getSetting
globalRolesForPermission = rolePermissionManager.getRolesForPermission
globalRolesForPrincipal = principalRoleManager.getRolesForPrincipal
SettingAsBoolean = {Allow: True, Deny: False, Unset: None, None: None}


class CacheEntry:
    pass


@zope.interface.provider(ISecurityPolicy)
class ZopeSecurityPolicy(ParanoidSecurityPolicy):

    def __init__(self, *args, **kw):
        ParanoidSecurityPolicy.__init__(self, *args, **kw)
        self._cache = {}

    def invalidate_cache(self):
        self._cache = {}

    def cache(self, parent):
        cache = self._cache.get(id(parent))
        if cache:
            cache = cache[0]
        else:
            cache = CacheEntry()
            self._cache[id(parent)] = cache, parent
        return cache

    def cached_decision(self, parent, principal, groups, permission):
        # Return the decision for a principal and permission

        cache = self.cache(parent)
        try:
            cache_decision = cache.decision
        except AttributeError:
            cache_decision = cache.decision = {}

        cache_decision_prin = cache_decision.get(principal)
        if not cache_decision_prin:
            cache_decision_prin = cache_decision[principal] = {}

        try:
            return cache_decision_prin[permission]
        except KeyError:
            pass

        # cache_decision_prin[permission] is the cached decision for a
        # principal and permission.

        decision = self.cached_prinper(parent, principal, groups, permission)
        if (decision is None) and groups:
            decision = self._group_based_cashed_prinper(parent, principal,
                                                        groups, permission)
        if decision is not None:
            cache_decision_prin[permission] = decision
            return decision

        roles = self.cached_roles(parent, permission)
        if roles:
            prin_roles = self.cached_principal_roles(parent, principal)
            if groups:
                prin_roles = self.cached_principal_roles_w_groups(
                    parent, principal, groups, prin_roles)
            for role, setting in prin_roles.items():
                if setting and (role in roles):
                    cache_decision_prin[permission] = decision = True
                    return decision

        cache_decision_prin[permission] = decision = False
        return decision

    def cached_prinper(self, parent, principal, groups, permission):
        # Compute the permission, if any, for the principal.
        cache = self.cache(parent)
        try:
            cache_prin = cache.prin
        except AttributeError:
            cache_prin = cache.prin = {}

        cache_prin_per = cache_prin.get(principal)
        if not cache_prin_per:
            cache_prin_per = cache_prin[principal] = {}

        try:
            return cache_prin_per[permission]
        except KeyError:
            pass

        if parent is None:
            prinper = SettingAsBoolean[
                globalPrincipalPermissionSetting(permission, principal, None)]
            cache_prin_per[permission] = prinper
            return prinper

        prinper = IPrincipalPermissionMap(parent, None)
        if prinper is not None:
            prinper = SettingAsBoolean[
                prinper.getSetting(permission, principal, None)]
            if prinper is not None:
                cache_prin_per[permission] = prinper
                return prinper

        parent = removeSecurityProxy(getattr(parent, '__parent__', None))
        prinper = self.cached_prinper(parent, principal, groups, permission)
        cache_prin_per[permission] = prinper
        return prinper

    def _group_based_cashed_prinper(self, parent, principal, groups,
                                    permission):
        denied = False
        for group_id, ggroups in groups:
            decision = self.cached_prinper(parent, group_id, ggroups,
                                           permission)
            if (decision is None) and ggroups:
                decision = self._group_based_cashed_prinper(
                    parent, group_id, ggroups, permission)

            if decision is None:
                continue

            if decision:
                return decision

            denied = True

        if denied:
            return False

        return None

    def cached_roles(self, parent, permission):
        cache = self.cache(parent)
        try:
            cache_roles = cache.roles
        except AttributeError:
            cache_roles = cache.roles = {}
        try:
            return cache_roles[permission]
        except KeyError:
            pass

        if parent is None:
            roles = {
                role: 1
                for (role, setting) in globalRolesForPermission(permission)
                if setting is Allow}
            cache_roles[permission] = roles
            return roles

        roles = self.cached_roles(
            removeSecurityProxy(getattr(parent, '__parent__', None)),
            permission)
        roleper = IRolePermissionMap(parent, None)
        if roleper:
            roles = roles.copy()
            for role, setting in roleper.getRolesForPermission(permission):
                if setting is Allow:
                    roles[role] = 1
                elif role in roles:
                    del roles[role]

        cache_roles[permission] = roles
        return roles

    def cached_principal_roles_w_groups(self, parent,
                                        principal, groups, prin_roles):
        denied = {}
        allowed = {}
        for group_id, ggroups in groups:
            group_roles = dict(self.cached_principal_roles(parent, group_id))
            if ggroups:
                group_roles = self.cached_principal_roles_w_groups(
                    parent, group_id, ggroups, group_roles)
            for role, setting in group_roles.items():
                if setting:
                    allowed[role] = setting
                else:
                    denied[role] = setting

        denied.update(allowed)
        denied.update(prin_roles)
        return denied

    def cached_principal_roles(self, parent, principal):
        cache = self.cache(parent)
        try:
            cache_principal_roles = cache.principal_roles
        except AttributeError:
            cache_principal_roles = cache.principal_roles = {}
        try:
            return cache_principal_roles[principal]
        except KeyError:
            pass

        if parent is None:
            roles = {
                role: SettingAsBoolean[setting]
                for (role, setting) in globalRolesForPrincipal(principal)}
            roles['zope.Anonymous'] = True  # Everybody has Anonymous
            cache_principal_roles[principal] = roles
            return roles

        roles = self.cached_principal_roles(
            removeSecurityProxy(getattr(parent, '__parent__', None)),
            principal)

        prinrole = IPrincipalRoleMap(parent, None)
        if prinrole:
            roles = roles.copy()
            for role, setting in prinrole.getRolesForPrincipal(principal):
                roles[role] = SettingAsBoolean[setting]

        cache_principal_roles[principal] = roles
        return roles

    def checkPermission(self, permission, object):
        if permission is CheckerPublic:
            return True

        object = removeSecurityProxy(object)
        seen = {}
        for participation in self.participations:
            principal = participation.principal
            if principal is system_user:
                continue  # always allow system_user

            if principal.id in seen:
                continue

            if not self.cached_decision(
                object, principal.id, self._groupsFor(principal), permission,
            ):
                return False

            seen[principal.id] = 1

        return True

    def _findGroupsFor(self, principal, getPrincipal, seen):
        result = []
        for group_id in getattr(principal, 'groups', ()):
            if group_id in seen:
                # Dang, we have a cycle.  We don't want to
                # raise an exception here (or do we), so we'll skip it
                continue
            seen.append(group_id)

            try:
                group = getPrincipal(group_id)
            except PrincipalLookupError:
                # It's bad if we have an undefined principal,
                # but we don't want to fail here.  But we won't
                # honor any grants for the group. We'll just skip it.
                continue

            result.append((group_id,
                           self._findGroupsFor(group, getPrincipal, seen)))
            seen.pop()

        return tuple(result)

    def _groupsFor(self, principal):
        groups = self._cache.get(principal.id)
        if groups is None:
            groups = getattr(principal, 'groups', ())
            if groups:
                getPrincipal = getUtility(IAuthentication).getPrincipal
                groups = self._findGroupsFor(principal, getPrincipal, [])
            else:
                groups = ()

            self._cache[principal.id] = groups

        return groups


def settingsForObject(ob):
    """Analysis tool to show all of the grants to a process
    """
    result = []
    while ob is not None:
        data = {}
        result.append((getattr(ob, '__name__', '(no name)'), data))

        principalPermissions = IPrincipalPermissionMap(ob, None)
        if principalPermissions is not None:
            settings = principalPermissions.getPrincipalsAndPermissions()
            data['principalPermissions'] = [
                {'principal': pr, 'permission': p, 'setting': s}
                for (p, pr, s) in sorted(settings)]

        principalRoles = IPrincipalRoleMap(ob, None)
        if principalRoles is not None:
            settings = principalRoles.getPrincipalsAndRoles()
            data['principalRoles'] = [
                {'principal': p, 'role': r, 'setting': s}
                for (r, p, s) in sorted(settings)]

        rolePermissions = IRolePermissionMap(ob, None)
        if rolePermissions is not None:
            settings = rolePermissions.getRolesAndPermissions()
            data['rolePermissions'] = [
                {'permission': p, 'role': r, 'setting': s}
                for (p, r, s) in sorted(settings)]

        ob = getattr(ob, '__parent__', None)

    data = {}
    result.append(('global settings', data))

    settings = principalPermissionManager.getPrincipalsAndPermissions()
    data['principalPermissions'] = [
        {'principal': pr, 'permission': p, 'setting': s}
        for (p, pr, s) in sorted(settings)]

    settings = principalRoleManager.getPrincipalsAndRoles()
    data['principalRoles'] = [
        {'principal': p, 'role': r, 'setting': s}
        for (r, p, s) in sorted(settings)]

    settings = rolePermissionManager.getRolesAndPermissions()
    data['rolePermissions'] = [
        {'permission': p, 'role': r, 'setting': s}
        for (p, r, s) in sorted(settings)]

    return result
